use crate::StdErr;
use futures::stream::StreamExt;
use rand::seq::SliceRandom;
use rand::{prelude::ThreadRng, thread_rng};
use songbird::Songbird;
use std::cell::RefCell;
use std::sync::Arc;
use tokio::select;
use tokio::sync::{broadcast, mpsc};
use tracing::{event, instrument, Level};
use twilight_gateway::{Cluster, Event};
use twilight_http::Client;
use twilight_model::{
    channel::message::Message,
    gateway::{
        payload::outgoing::update_presence::UpdatePresencePayload,
        presence::{ActivityType, MinimalActivity, Status},
        Intents,
    },
    id::{marker::UserMarker, Id},
};

thread_local!(static RNG: RefCell<ThreadRng> = RefCell::new(thread_rng()));

const GREETINGS: &'static [&'static str] = &[
    "You rung?",
    "Hello there",
    "Right back atcha",
    "Yes?",
    "Wow, rude",
    "Go stick your head in a pig",
];

#[derive(Debug)]
pub struct Call {
    pub guild_id: u64,
    pub channel_id: u64,
    pub done: mpsc::Receiver<()>,
}

struct State {
    songbird: Songbird,
    client: Client,
}

pub async fn run_discord(
    token: &str,
    mut calls: mpsc::Receiver<Call>,
    mut shutdown: broadcast::Receiver<()>,
    _done: mpsc::Sender<()>,
) -> StdErr<()> {
    event!(Level::INFO, "Starting...");
    let (cluster, mut events) = Cluster::builder(
        token.to_owned(),
        Intents::GUILD_MESSAGES | Intents::MESSAGE_CONTENT | Intents::GUILD_VOICE_STATES,
    )
    .presence(UpdatePresencePayload::new(
        vec![MinimalActivity {
            kind: ActivityType::Listening,
            name: "Your INVITEs".to_owned(),
            url: None,
        }
        .into()],
        false,
        None,
        Status::Online,
    )?)
    .build()
    .await?;
    let cluster = Arc::new(cluster);
    let client = Client::new(token.to_owned());
    let me = client.current_user().exec().await?.model().await?.id;
    let sb = Songbird::twilight(cluster.clone(), me);
    let state = Arc::new(State {
        songbird: sb,
        client,
    });
    let cluster_spawn = cluster.clone();
    tokio::spawn(async move { cluster_spawn.up().await });

    loop {
        select!(
            Some((shard_id, ev)) = events.next() => {
                state.songbird.process(&ev).await;
                let state_ = state.clone();
                tokio::spawn(async move {
                    if let Err(e) = handle_event(ev, shard_id, state_, me).await {
                        event!(Level::ERROR, err=?e, "Error handling discord event");
                    }
                });
            },
            Some(call) = calls.recv() => {
                let state_ = state.clone();
                tokio::spawn(handle_call(call, state_));
            },
            _ = shutdown.recv() => {
                event!(Level::INFO, "Shutting down...");
                cluster.down();
                break;
            },

        );
    }
    event!(Level::INFO, "Done");
    Ok(())
}

#[instrument(skip(call, state))]
async fn handle_call(mut call: Call, state: Arc<State>) {
    let (_handle, success) = state.songbird.join(call.guild_id, call.channel_id).await;
    match success {
        Ok(()) => event!(Level::INFO, %call.guild_id, %call.channel_id, "Joined channel"),
        Err(err) => {
            event!(Level::ERROR, %call.guild_id, %call.channel_id, %err, "Error joining channel")
        }
    }
    let _ = call.done.recv().await;
    match state.songbird.leave(call.guild_id).await {
        Ok(()) => event!(Level::INFO, %call.guild_id, %call.channel_id, "Left channel"),
        Err(err) => {
            event!(Level::INFO, %call.guild_id, %call.channel_id, %err, "Error leaving channel")
        }
    }
}

fn mentions(msg: &Message, us: Id<UserMarker>) -> bool {
    msg.mentions.iter().filter(|m| m.id == us).next().is_some()
}

#[instrument(skip(state))]
async fn handle_event(
    ev: Event,
    shard_id: u64,
    state: Arc<State>,
    me: Id<UserMarker>,
) -> StdErr<()> {
    match ev {
        Event::MessageCreate(msg) => {
            if mentions(&msg.0, me) {
                let greet = RNG.with(|rng| GREETINGS.choose(&mut *rng.borrow_mut()).unwrap());
                state
                    .client
                    .create_message(msg.channel_id)
                    .content(greet)?
                    .exec()
                    .await?;
            }
        }
        Event::ShardConnected(_) => event!(Level::INFO, "Shard connected"),
        _ => {}
    }
    Ok(())
}
